/*
 * binary_io.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

//#define DEBUG_WRITE

#include "binary_io.hpp"

#include <iostream>
#include <stdexcept>
using namespace std;

signed char BinReadByte(istream &str)
{
	return str.get();
}

short int BinReadWord(istream &str)
{
	signed short int val;
	str.read(reinterpret_cast<char*>(&val), 2);
	return val;
}

int BinReadLong(istream &str)
{
	int val;
	str.read(reinterpret_cast<char*>(&val), 4);
	return val;
}

unsigned char BinReadUbyte(istream &str)
{
	return str.get();
}

unsigned short int BinReadUword(istream &str)
{
	unsigned short int val;
	str.read(reinterpret_cast<char*>(&val), 2);
	return val;
}

unsigned int BinReadUlong(istream &str)
{
	unsigned int val;
	str.read(reinterpret_cast<char*>(&val), 4);
	return val;
}

float BinReadFloat(istream &str)
{
	float val;
	str.read(reinterpret_cast<char*>(&val), 4);
	return val;
}


void BinWriteByte(ostream &str, signed char val)
{
#ifdef DEBUG_WRITE
	cout << "writing byte: " << int(val) << endl;
#endif
	str.put(val);
}

void BinWriteWord(ostream &str, short int val)
{
#ifdef DEBUG_WRITE
	cout << "writing word: " << val << endl;
#endif
	str.write(reinterpret_cast<char*>(&val), 2);
}

void BinWriteLong(ostream &str, int val)
{
#ifdef DEBUG_WRITE
	cout << "writing long: " << val << endl;
#endif
	str.write(reinterpret_cast<char*>(&val), 4);
}

void BinWriteUbyte(ostream &str, unsigned char val)
{
#ifdef DEBUG_WRITE
	cout << "writing unsigned byte: " << int(val) << endl;
#endif
	str.put(val);
}

void BinWriteUword(ostream &str, unsigned short int val)
{
#ifdef DEBUG_WRITE
	cout << "writing unsigned word: " << val << endl;
#endif
	str.write(reinterpret_cast<char*>(&val), 2);
}

void BinWriteUlong(ostream &str, unsigned int val)
{
#ifdef DEBUG_WRITE
	cout << "writing unsigned long: " << val << endl;
#endif
	str.write(reinterpret_cast<char*>(&val), 4);
}

void BinWriteFloat(ostream &str, float val)
{
#ifdef DEBUG_WRITE
	cout << "writing float: " << val << endl;
#endif
	str.write(reinterpret_cast<char*>(&val), 4);
}


//
// strings
//

string BinReadString(istream &str)
{
	string s;
	int size = BinReadInt(str);
	s.resize(size);
	for (string::iterator it = s.begin(); it != s.end(); ++it) {
		*it = str.get();
	}
	return s;
}

void BinWriteString(ostream &str, const string &s)
{
	BinWriteInt(str, s.size());
#ifdef DEBUG_WRITE
	cout << "writing string data: " << s << endl;
#endif
	for (string::const_iterator it = s.begin(); it != s.end(); ++it) {
		str.put(*it);
	}
}


//
// variable-size ints
//

int BinReadInt(istream &str)
{
	int val;
	bool sign;
	char c;
	str.get(c);
	// bit 7 = extend bit
	// bit 6 = sign bit
	// bits 5-0 = least significant bits
	val = (c & 0x3f);
	sign = (c & 0x40);
	if (c & 0x80) {
		// 2nd byte
		str.get(c);
		val |= ((c & 0x7f) << 6);
		if (c & 0x80) {
			// 3rd byte
			str.get(c);
			val |= ((c & 0x7f) << 13);
			if (c & 0x80) {
				// 4th byte
				str.get(c);
				val |= (static_cast<unsigned char>(c) << 20);
			} 
		}
	}
	if (sign) {
		val = -val-1;
	}
	return val;
}

void BinWriteInt(ostream &str, int val)
{
#ifdef DEBUG_WRITE
	cout << "writing int: " << val << endl;
	if (!str) {
		cout << "WARNING: stream has gone invalid" << endl;
	}
#endif
	
	int sign;
	if (val >= 0) {
		sign = 0;
	} else {
		sign = 0x40;
		val = -val-1;
	}

	if (val > 268435455) {
		throw domain_error("BinWrite(int) out of range");
	}

	if (val <= 63) {
		str.put(sign | val);
	} else {
		str.put(0x80 | sign | (val & 0x3f));
		if (val <= 8191) {
			str.put(val >> 6);
		} else {
			str.put(0x80 | ((val >> 6) & 0x7f));
			if (val <= 1048575) {
				str.put(val >> 13);
			} else {
				str.put(0x80 | ((val >> 13) & 0x7f));
				str.put(val >> 20);
			}
		}
	}
}	


