/*
 * knight_task.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "action.hpp"
#include "anim.hpp"
#include "cfg.hpp"
#include "control.hpp"
#include "dungeon_view.hpp"
#include "item.hpp"
#include "knight.hpp"
#include "knight_task.hpp"
#include "player.hpp"
#include "task_manager.hpp"

void KnightTask::execute(TaskManager &tm)
{
	// This routine calls Player::computeAvailableControls at regular
	// intervals. It might be better to call computeAvailableControls
	// only when needed (eg after motion, or after any change to
	// items, tiles etc in a square adjacent to the knight), ie in an
	// event driven way. That would be more efficient, but also more
	// complicated than what we are doing now.

	// This routine is also responsible for executing controls
	// selected by the player (information on the selected controls is
	// read from DungeonView).

	// Also: we do crossbow loading here.

	player.computeAvailableControls();
	
	shared_ptr<Knight> knight = player.getKnight();
	if (!knight || !knight->getMap()) return;  // knight not in map at present

	const int gvt = tm.getGVT();

	// Process controls.
	DungeonView &dview (player.getDungeonView());
	const Control * ctrl = dview.readController();
	if (!knight->stunned()) {
		dview.setMenuHighlight(0);  // clear the menu highlight (if not stunned)

		if (ctrl && ctrl->getAction() ) {
			// first clear any crossbow animation...
			// (if we don't do this then e.g. knight will still be on AF_XBOW_LOAD when he starts moving.)
			if (knight->getAnimFrame() == AF_XBOW_LOAD) knight->setAnimFrame(AF_XBOW, 0);
			
			// execute selected control
			const ItemType *item_type;
			weak_ptr<Tile> tile;
			MapCoord tile_mc;
			player.getControlInfo(ctrl, item_type, tile, tile_mc);
			ActionData ad;
			ad.setActor(knight, true);
			shared_ptr<Tile> tile_lock = tile.lock();
			if (item_type) ad.setItem(0, MapCoord(), item_type);
			if (tile_lock) ad.setTile(knight->getMap(), tile_mc, tile_lock);
			ctrl->getAction()->execute(ad);
			
			// also set menu highlight to that control
			dview.setMenuHighlight(ctrl);
		}
	}

	// Work out the crossbow loading.
	const ItemType *held_item = knight->getItemInHand();
	bool can_load_xbow = knight->canAct()
		&& held_item && held_item->canLoad()
		&& held_item->getAmmo() && knight->getNumCarried(*held_item->getAmmo()) > 0
		&& !knight->isApproaching();
	if (!can_load_xbow) {
		xbow_load_time = 0;
		xbow_action_timer = 0;
	} else if (xbow_load_time == 0) {
		int reload_time = held_item->getReloadTime();
		if (knight->hasQuickness()) reload_time /= 2;
		xbow_load_time = gvt + reload_time;
		xbow_action_timer = 0;
	} else if (gvt >= xbow_load_time) {
		knight->setItemInHand(held_item->getLoaded());
		knight->rmFromBackpack(*held_item->getAmmo(), 1);
		xbow_load_time = 0;
		xbow_action_timer = 0;
	}
	held_item = knight->getItemInHand();

	// Play clicking sound while crossbow is loading (reload_action).
	if (held_item && xbow_load_time != 0 && held_item->getReloadAction()
	&& gvt >= xbow_action_timer + held_item->getReloadActionTime()) {
		xbow_action_timer = gvt;
		ActionData ad;
		ad.setActor(knight, false);
		held_item->getReloadAction()->execute(ad);
	}	

	// Set anim frame for crossbow loading if necessary
	int af = knight->getAnimFrame();
	if (af == AF_NORMAL || af == AF_XBOW || af == AF_XBOW_LOAD) {
		if (held_item) {
			if (held_item->canLoad() && xbow_load_time != 0) {
				knight->setAnimFrame(AF_XBOW_LOAD, 0);
			} else if (held_item->canShoot() || held_item->canLoad()) {
				knight->setAnimFrame(AF_XBOW, 0);
			} else {
				knight->setAnimFrame(AF_NORMAL, 0);
			}
		} else {
			knight->setAnimFrame(AF_NORMAL, 0);
		}
	}
	
	// Set up for next time. We keep polling at
	// 'control_poll_interval'. 
	tm.addTask(shared_from_this(), TP_NORMAL, gvt + CfgInt("control_poll_interval"));
}

