/*
 * monster_definitions.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "anim.hpp"
#include "cfg.hpp"
#include "dungeon_map.hpp"
#include "item.hpp"
#include "mediator.hpp"
#include "misc/rng.hpp"
#include "monster_definitions.hpp"
#include "monster_manager.hpp"
#include "monster_support.hpp"
#include "special_tiles.hpp"
#include "task.hpp"
#include "task_manager.hpp"
#include "tile.hpp"

#include "kconfig/random_int.hpp"
using namespace KConfig;

//
// vampire bats
//

class VampireBatAI : public Task {
public:
	explicit VampireBatAI(weak_ptr<VampireBat> v) : vbat(v), bit_last_time(false) { }
	virtual void execute(TaskManager &tm);
private:
	weak_ptr<VampireBat> vbat;
	bool bit_last_time;
};

namespace {
	struct Always {
		bool operator()(shared_ptr<Knight> kt) const { return true; }
	};

	struct BatCanEnter {
		bool operator()(DungeonMap &dmap, const MapCoord &mc) {
			return (dmap.getAccess(mc, H_FLYING) == A_CLEAR);
		}
	};

	void ReplaceTask(TaskManager &tm, shared_ptr<Task> task, Monster &mon)
	{
		bool can_act = mon.canAct();
		int cannot_act_until;
		if (!can_act) {
			bool known;
			mon.cannotActUntil(known, cannot_act_until);
			if (!known) can_act = true;   // Treat unknown case as if we can act (ie. just wait a preset delay time).
		}
		if (can_act) {
			// Monster did not do anything. Wait for a preset delay time.
			tm.addTask(task, TP_LOW, tm.getGVT() + CfgInt("monster_wait_time"));
		} else {
			// Wait until the monster's current action finishes.
			tm.addTask(task, TP_LOW, cannot_act_until + 1);
		}
	}
}

void VampireBatAI::execute(TaskManager &tm)
{
	shared_ptr<VampireBat> bat = vbat.lock();
	if (!bat || !bat->getMap()) return;  // the bat has died

	// These vars store whether we should move (and in which dir) and whether we should bite
	pair<MapDirection,bool> p;
	p.second = false;
	bool bite = false;
	
	// Find a target.
	shared_ptr<Knight> target = FindClosestKnight(bat, Always());

	// Choose a direction to move in.
	if (bat->getRunAwayFlag() && target) {
		// "Run away flag" means we must move in the direction that the target is facing
		// (or a standard "run away" direction if that is not possible)
		MapDirection dir = target->getFacing();
		if (BatCanEnter()(*bat->getMap(), DisplaceCoord(bat->getPos(), dir))) {
			p.first = dir;
			p.second = true;
		} else {
			p = ChooseDirection(bat, target->getPos(), true, BatCanEnter());
		}
	} else if (target && target->getPos() == bat->getPos() && !bit_last_time) {
		// We are over the top of a target, and we have not bit him yet
		// Bite the target
		bite = true;
	} else if (target || !g_rng.getBool(CfgInt("monster_wait_chance")/100.0f)) {
		// Otherwise, do a standard move towards the target (if any).
		// (Special rule: If there is no target then there is a "monster_wait_chance"
		// percentage chance of doing nothing, as for zombies.)
		p = ChooseDirection(bat, target? target->getPos() : MapCoord(), false, BatCanEnter());
	}

	// Move if can act
	if (bat->canAct()) {
		bat->clearRunAwayFlag();
		if (p.second) {
			bit_last_time = false;
			bat->setFacing(p.first);
			bat->move(MT_MOVE);
		} else if (bite) {
			bit_last_time = true;
			bat->bite();
		}
	}

	// Now replace the task
	ReplaceTask(tm, shared_from_this(), *bat);
}

shared_ptr<Monster> VampireBatMonsterType::makeMonster(MonsterManager &mm, TaskManager &tm)
	const
{
	const int h = health ? health->get() : 1; 
	shared_ptr<VampireBat> vbat(new VampireBat(mm, *this, h, anim, speed));
	shared_ptr<Task> ai(new VampireBatAI(vbat));
	tm.addTask(ai, TP_LOW, tm.getGVT()+1);
	return vbat;
}


void VampireBat::damage(int amount, int su)
{
	// run vampire bat hook (usually plays "screech" sound effect)
	// -- Only want this if the bat was not killed.
	shared_ptr<VampireBat> self(static_pointer_cast<VampireBat>(shared_from_this()));
	if (amount < getHealth()) {
		Mediator::instance().runHook("HOOK_BAT", self);
	} else {
		Mediator::instance().runHook("HOOK_CREATURE_SQUELCH", self);
	}
	
	// Vampire bats will run away after they get hit.
	// Also: Vampire bats are immune to being "stunned" by weapon impacts.
	Creature::damage(amount, -1);
	run_away_flag = true;
}

void VampireBat::bite()
{
	if (!canAct()) return;
	DungeonMap *dmap = getMap();
	if (!dmap) return;
	const int gvt = Mediator::instance().getGVT();

	// strike the target
	shared_ptr<Creature> target = dmap->getTargetCreature(*this, false);
	if (target) {
		const int stun_until = (mtype.stun? mtype.stun->get() : 0) + gvt;
		target->damage(mtype.dmg, stun_until);
	}

	// set my anim frame, and stun myself.
	const int wait_until = gvt + CfgInt("melee_delay_time");
	setAnimFrame(AF_IMPACT, wait_until);
	stunUntil(wait_until);
}


//
// zombies
//

class ZombieAI : public Task {
public:
	ZombieAI(const MonsterManager &mm, weak_ptr<Monster> z) : mmgr(mm), zombie(z) { }
	virtual void execute(TaskManager &tm);
private:
	const MonsterManager &mmgr;
	weak_ptr<Monster> zombie;
};

namespace {
	// some predicates:
	
	struct IsVisible {
		bool operator()(shared_ptr<Knight> kt) const {
			return kt->isVisible();
		}
	};

	struct VisibleAndCarrying {
		explicit VisibleAndCarrying(const ItemType *it_) : it(it_) { }
		bool operator()(shared_ptr<Knight> kt) const {
			return kt->getItemInHand() == it && kt->isVisible();
		}
		const ItemType *it;
	};

	struct ZombieCanWalk {
		explicit ZombieCanWalk(const MonsterManager &m) : mm(m) { }
		const MonsterManager &mm;
		bool operator()(DungeonMap &dmap, const MapCoord &mc) const {
			// If the access is not A_CLEAR then we may not walk into this square
			if (dmap.getAccess(mc, H_WALKING) != A_CLEAR) return false;
			
			// If a tile is on the "avoid" list then we may not walk into this square
			vector<shared_ptr<Tile> > tiles;
			dmap.getTiles(mc, tiles);
			for (vector<shared_ptr<Tile> >::iterator it = tiles.begin(); it != tiles.end();
			++it) {
				if (find(mm.getZombieAvoid().begin(), mm.getZombieAvoid().end(), *it)
				!= mm.getZombieAvoid().end()) {
					return false;
				}
			}

			return true;
		}
	};

	struct ZombieCanFight {
		explicit ZombieCanFight(const MonsterManager &m) : mm(m) { }
		const MonsterManager &mm;
		bool operator()(DungeonMap &dmap, const MapCoord &mc) const {
			
			// A zombie can always attack a knight (assuming we're not afraid of him)
			// Note: this means a zombie will be able to attack an invisible knight if it
			// is next to such a knight. But zombies will not *target* invisible knights
			// from a distance.
			if (KnightAt(dmap, mc, mm.getZombieFear())) return true;

			// A zombie can fight a "bear trap" tile
			if (dmap.getItem(mc) && &dmap.getItem(mc)->getType() == mm.getZombieHit()) {
				return true;
			}
			
			// A zombie can also try to smash furniture tiles (as long as they're
			// not doors).
			vector<shared_ptr<Tile> > tiles;
			dmap.getTiles(mc, tiles);
			for (vector<shared_ptr<Tile> >::iterator it = tiles.begin(); it != tiles.end();
			++it) {
				if ((*it)->destructible() && !dynamic_cast<Door*>(it->get())) {
					return true;
				}
			}

			return false;
		}
	};

	struct ZombieCanMove {
		explicit ZombieCanMove(const MonsterManager &m) : mm(m) { }
		bool operator()(DungeonMap &dmap, const MapCoord &mc) const {
			const ZombieCanWalk zcw(mm);
			const ZombieCanFight zcf(mm);
			return zcw(dmap,mc) || zcf(dmap,mc);
		}
		const MonsterManager &mm;
	};
}

void ZombieAI::execute(TaskManager &tm)
{
	shared_ptr<Monster> zom = zombie.lock();
	if (!zom || !zom->getMap()) return;  // our zombie appears to have died.

	pair<MapDirection,bool> p;
	p.second = false;
	
	// Find a target (or something to run away from!)
	shared_ptr<Knight> target = FindClosestKnight(zom,
			VisibleAndCarrying(mmgr.getZombieFear()));
	bool run_away;
	if (target) {
		run_away = true;
	} else {
		target = FindClosestKnight(zom, IsVisible());
		run_away = false;
	}
	
	// Choose a direction to move in:
	// (If there is no target, then there is a chance that the monster will stay
	// where it is and do nothing, rather than randomly walking about.)
	if (!(!target && g_rng.getBool(CfgInt("monster_wait_chance")/100.0f))) {
		p = ChooseDirection(zom, target? target->getPos() : MapCoord(), run_away,
							ZombieCanMove(mmgr));
	}

	// Move (if we can act)
	if (zom->canAct()) {

		if (p.second) {
			// A direction was chosen above. Turn to face this direction
			zom->setFacing(p.first);
			
			// Either fight or walk, depending on what's in the tile ahead.
			MapCoord sq_ahead = DisplaceCoord(zom->getPos(), zom->getFacing());
			const ZombieCanFight zcf(mmgr);
			const ZombieCanWalk zcw(mmgr);
			if (zcf(*zom->getMap(), sq_ahead)) {
				zom->swing();
			} else if (zcw(*zom->getMap(), sq_ahead)) {
				zom->move(MT_MOVE);
				// Play a moo sound 1 in every 20 zombie moves
				if (g_rng.getBool(0.05f)) {
					Mediator::instance().runHook("HOOK_ZOMBIE", zom);
				}
			}
		} else {
			// We have chosen to stay where we are. But we should at least
			// turn to face the player (this looks a bit better).
			if (target) {
				zom->setFacing(DirectionFromTo(zom->getPos(), target->getPos()));
			} else {
				zom->setFacing(MapDirection(g_rng.getInt(0,4)));
			}
		}
	}

	// Now wait for an appropriate time before making the next move. 
	ReplaceTask(tm, shared_from_this(), *zom);
}

shared_ptr<Monster> ZombieMonsterType::makeMonster(MonsterManager &mm, TaskManager &tm) const
{
	const int h = health ? health->get() : 1;
	shared_ptr<Monster> zombie(new Zombie(mm, *this, h, weapon, anim, speed));
	zombie->setFacing(MapDirection(g_rng.getInt(0,4))); // random initial facing
	shared_ptr<Task> ai(new ZombieAI(mm, zombie));
	tm.addTask(ai, TP_LOW, tm.getGVT()+1);
	return zombie;
}

//
// This routine makes zombies 'recoil' when they're hit.
// Also it runs the hooks (usually sound effects).

void Zombie::damage(int amount, int stun_until)
{
	shared_ptr<Zombie> self(static_pointer_cast<Zombie>(shared_from_this()));
	Mediator::instance().runHook("HOOK_ZOMBIE", self);
	if (amount >= getHealth()) {
		Mediator::instance().runHook("HOOK_CREATURE_SQUELCH", self);
	}
	Creature::damage(amount, stun_until);
	setAnimFrame(AF_PARRY,
				 stun_until != -1 ? stun_until
				                  : (Mediator::instance().getGVT() + CfgInt("zombie_damage_delay")));
}
